///////////////////////////////////////////////////////////////////////////////
//  untSOAPStub:                                                             //
//  Generic SOAP stub. Parse an XML SOAP method call, execute the requested  //
//  method, package any result in an XML SOAP response and return it.        //
//  Do not instantiate TCustomSOAPStub directly. Descend from it and         //
//  implement the abstract ProcessRequest method.                            //
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//  Written for the January/February 2001 issue of                           //
//  "uk-bug developer's magazine"                                            //
//  by Rob Bracken                                                           //
//  Bracken Software Limited                                                 //
//  01275 843215                                                             //
///////////////////////////////////////////////////////////////////////////////

unit untSOAPStub;

interface

uses
  XDOM;

type
  TCustomSOAPStub = class(TObject)
  private

    FXMLParser: TXmlToDomParser;
    FDomImpl: TDomImplementation;
    FDomDoc: TDOMDocument;
    FURI: string;               // The URI specified in the SOAPAction attribute

  protected

    FServerObject: TObject;

    function ProcessRequest(XMLNode: TDOMNode; MethodName: string): string;
                  virtual; abstract; // Must be overridden in descendant classes
    property URI: string write FURI;

  public

    constructor Create(ServerObject: TObject); virtual;
    destructor Destroy; override;

    function Execute(Request: string): string;

  published

  end;


implementation

{ TCustomSOAPStub }

constructor TCustomSOAPStub.Create(ServerObject: TObject);
begin
  // Create the XML parsing components
  FXMLParser := TXmlToDomParser.Create(nil);
  FDomImpl := TDomImplementation.Create(nil);
  FXMLParser.DOMImpl := FDomImpl;
  FServerObject := ServerObject;
end;

destructor TCustomSOAPStub.Destroy;
begin
  // Free the XML parsing components
  FXMLParser.Free;
  FDomImpl.Free;
  inherited;
end;

function TCustomSOAPStub.Execute(Request: string): string;
{ Parse the XML request and execute method calls }

  function GetNextXMLElement(XMLNode: TDOMNode): TDOMNode;
  { Find the _next_ XML element node at this level (NodeType = ntElement_Node) }
  begin
    Result := XMLNode.NextSibling;
    while (Result.NodeType <> ntElement_Node) and (Result <> nil) do
      Result := Result.NextSibling;
  end;

  function FindNode(XMLNode: TDOMNode; sValueToFind: string): TDOMNode;
  { Find the first XML element at this level that has the supplied value,
    starting with the supplied node. (An XML element is a node with a NodeType
    of ntElement_Node)
    Note that we actually look at the NodeName property, rather than the
    NodeValue. }
  begin

    if XMLNode.NodeType <> ntElement_Node then
    begin
      Result := GetNextXMLElement(XMLNode);
    end
    else
    begin
      Result := XMLNode;
    end;
    // Find the first child that has a matching value
    while (Result.NodeName <> sValueToFind) and (Result <> nil) do
      Result := GetNextXMLElement(XMLNode);
  end;

  function GetMethodName(SOAPMethodNode: TDOMNode): string;
  { Extract the method name from the node name. The node name is in the form
    "ns:method" where ns = the XML namespace and method = the method name,
    so we need to extract everything after the colon. }
  var
    nColonPos: integer;
  begin
    nColonPos := Pos(':',SOAPMethodNode.NodeName);
    Result := Copy(SOAPMethodNode.NodeName,nColonPos+1,
                   length(SOAPMethodNode.NodeName)-nColonPos);
  end;

var
  SOAPEnvelopeNode, SOAPBodyNode, SOAPMethodNode: TDomNode;
begin

  // Parse the XML and load it into a DOM document tree
  FDomDoc := FXMLParser.stringToDom(Request);

  // Find the SOAP envelope node (will be a child of the DOM document node)
  SOAPEnvelopeNode := FindNode(FDomDoc.FirstChild, 'SOAP-ENV:Envelope');

  // Find the SOAP Body node (will be a child of the envelope node)
  SOAPBodyNode := FindNode(SOAPEnvelopeNode.FirstChild, 'SOAP-ENV:Body');

  // Method calls will be children of the SOAP Body node.
  // For simplicity, we've assumed that all the method calls are for us, so we
  // don't check the namespace. To check the namespace, find the "xmlns"
  // attribute of the method node and compare its value.
  SOAPMethodNode := SOAPBodyNode.FirstChild;
  while Assigned(SOAPMethodNode) do
  begin
    // ProcessRequest takes a SOAP method node, performs the action and returns
    // the result as XML. Actions which don't return anything will return a
    // blank string. If more than one method is called, the return strings are
    // concatenated together.
    Result := Result + ProcessRequest(SOAPMethodNode, GetMethodName(SOAPMethodNode));
    SOAPMethodNode := SOAPMethodNode.NextSibling;
    while Assigned(SOAPMethodNode) and
          (SOAPMethodNode.NodeType <> ntElement_Node) do
      SOAPMethodNode := SOAPMethodNode.NextSibling;           // Skip any text or comment nodes
  end;

  // Wrap the result in a SOAP envelope.
  Result := '<SOAP-ENV:Envelope' +
            ' xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/"' +
            ' SOAP-ENV:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">' +
            '<SOAP-ENV:Body>' +
            Result +
            '</SOAP-ENV:Body>' +
            '</SOAP-ENV:Envelope>';

end;

end.
